# Created on 09/12/2021
from polynome import *
from inverse_mult import *
from itertools import combinations


# Global Variables
# Parameters
INDEX_MAX = 20
ERROR_MAX = 9


def get_possibilities(l_og: list, index: int):
	''' Génère une liste qui contient des listes de points

	Args:
		l_og (list): liste de tous les points
		index (int): index à partir duquel les points sont forcément juste

	Returns:
		[list]: Liste de toutes les possibilitées de points
	'''	
	l = l_og.copy()
	# Transforme the list in list of tupples with their indexes
	for x, _ in enumerate(l):
		l[x] = (x, l[x])

	l_fixe = l[index:] # Liste contenant aucune erreur
	l_posi = l[:index] # Liste contenant des erreurs, 

	p = list(combinations(l_posi, 2))

	for x, y in enumerate(p):
		p[x] = list(y)
		p[x].extend(l_fixe)

	return p

def lagrange_compute(l: list) -> polynome:
	''' Retourne le polynome de Lagrange pour une 
	liste de points données

	Args:
		l (list): La liste de points

	Returns:
		polynome: Le polynome de Lagrange correspondant
	'''	

	poly_l = [0] * len(l)
	poly = polynome(poly_l)
	for x_i, y_i in enumerate(l):
		poly_n = polynome([1])
		for x_k, y_k in enumerate(l):
			if y_k[0] != y_i[0]:
				d = inverse_mult((y_i[0]-y_k[0]) % poly_n.prime_mod, poly_n.prime_mod)
				poly_n = poly_n.mul(polynome([-y_k[0] * d % poly_n.prime_mod, d % poly_n.prime_mod]))
		poly_n = poly_n.mul(polynome([y_i[1] % poly_n.prime_mod]))
		poly = poly.add(poly_n)
	return poly


def check_nb_error(l: list, p: polynome) -> int:
	''' Compte le nombre d'erreure entre le polynome et les points de la liste

	Args:
		l (list): La liste de points à tester
		p (polynome): Le polynome avec lequel tester la liste

	Returns:
		int: [description]
	'''		

	nbr_error = 0
	for counter, item in enumerate(l):
		if (item % p.prime_mod) != (p.evaluate(counter) % p.prime_mod):
			nbr_error += 1
	return nbr_error


def reed_decode(l: list) -> polynome:
	''' Teste toutes les listes de points et retourne le polynome de lagrange 
	lorsque la liste à moins de 9 erreurs

	Args:
		l (list): Liste de liste de points

	Returns:
		polynome: Polynome de Lagrange
	'''	

	for points in get_possibilities(l, INDEX_MAX):
		err = check_nb_error(l, lagrange_compute(points))
		if ((err) <= ERROR_MAX):
			# ? Chercher un polynome avec moins de 9 erreurs ?
			# Possibilité d'optimiser en evitant de compute 2 fois le poly
			return lagrange_compute(points)




def correct_list(len_list: int, p: polynome) -> list:
	''' Corrige la liste reçue en utilisant le bon polynome de Lagrange

	Args:
		len_list (int): Liste avec des erreurs
		p (polynome): Le bon polynome de Lagrange

	Returns:
		list: La liste corrigée
	'''	
	
	corrected = []

	for x in range(len_list):
		corrected.append(p.evaluate(x))

	return corrected